/*
 * _globals.h
 *
 *  Created on: 15. aug. 2018
 *      Author: teig
 */

#ifndef GLOBALS_H_
    #define GLOBALS_H_

    #ifdef GLOBALS_H_ // To show that the below may also be defined in library space

        // BOOLEAN #include <stdbool.h> if C99
        // See http://www.teigfam.net/oyvind/home/technology/165-xc-code-examples/#bool
        typedef enum {false,true} bool; // 0,1 This typedef matches any integer-type type like long, int, unsigned, char, bool

        #define min(a,b) (((a)<(b))?(a):(b)) // combining min and max..
        #define max(a,b) (((a)>(b))?(a):(b)) // ..gives range or not in range
        //
        // Observe sklearn.preprocessing.MinMaxScaler as interesting, perhaps?
        // https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html
        // S_scaled = (X - X_min) / (X_max - X_min)

        #define abs(a) (((a)<0)?(-(a)):(a))

        #define t_swap(type,a,b) {type t = a; a = b; b = t;}

        #define NUM_ELEMENTS(array) (sizeof(array) / sizeof(array[0])) // Kernighan & Pike p22

        typedef signed int time32_t; // signed int (=signed) or unsigned int (=unsigned) both ok, as long as they are monotoneously increasing
                                     // XC/XMOS 100 MHz increment every 10 ns for max 2exp32 = 4294967296,
                                     // ie. divide by 100 mill = 42.9.. seconds

        typedef enum {led_on, led_off}     led_on_low_e; // 0 is led_on
        typedef enum {port_low, port_high} port_e;

        #define AFTER_32(a,b) ((a-b)>0)

        #ifdef DO_ASSERTS
            #define ASSERT_DELAY_32(d) do {if (d > INT_MAX) fail("Overflow");} while (0) // Needs <so646.h<, <limits.h> and <xassert.h>
            // INT_MAX is 2147483647 is what fits into 31 bits or last value before a signed 32 bits wraps around
        #else
            #define ASSERT_DELAY_32(d)
        #endif

        #define NOW_32(tmr,time) do {tmr :> time;} while(0) // A tick is 10ns
        // Programming XC on XMOS Devices (Douglas Watt)
        //     If the delay between the two input values fits in 31 bits, timerafter is guaranteed to behave correctly,
        //     otherwise it may behave incorrectly due to overlow or underflow. This means that a timer can be used to
        //     measure up to a total of 2exp31 / (100 mill) = 21s.

    #endif

    #define IS_MYTARGET_VOID               0
    #define IS_MYTARGET_STARTKIT           1
    #define IS_MYTARGET_XCORE_200_EXPLORER 2
    #define IS_MYTARGET_XCORE_XA_MODULE    3 // Not able to flash it
    #define IS_MYTARGET_MIC_ARRAY          4

    #if (MYTARGET==MIC_ARRAY)
        #define IS_MYTARGET IS_MYTARGET_MIC_ARRAY
        #define NUM_LEDS       13
        #define NUM_LEDS_08_12 (NUM_LEDS-8)
    #else
        #error NO TARGET DEFINED
    #endif

    #ifndef DISPLAY_FAST_DARK
        #error define DISPLAY_FAST_DARK in makefile
    #else
        #if (DISPLAY_FAST_DARK>1)
            #error not defined
        #endif
    #endif

    #define MAX_SIZEOF_NONBLOCKING_STREAMING_CHAN (8) // 2 words * 4 bytes

    #ifndef USE_XSCOPE
        #error USE_XSCOPE not defined in makefile
    #elif (USE_XSCOPE == 0)
        #define DO_XSCOPE_INT(name,val,use_at)
        #define DO_XSCOPE_START(name,use_at)
        #define DO_XSCOPE_STOP(name,use_at)
    #elif (USE_XSCOPE == 1)
        // Needs #include <xscope.h> // xScope logging

        #define USE_XSCOPE_AT 1 // 0 no place, other (below)
        //                    #                                   config.xscope
        #define XSCOPE_AT_1   1 // In mic_sample in mics_handler. CONTINUOUS  INT
        #define XSCOPE_AT_2   2 // In logger.                     STARTSTOP  UINT
        //
        // If config.xscope type="STARTSTOP" then all "out" must be framed by
        // DO_XSCOPE_START/DO_XSCOPE_STOP. If not, this:
        //   xSCOPE warning ... START/STOP event 0 not called with xscope_start()/xscope_stop()

        // This takes about 0.2 us, but ROOT_DELAY_UNTIL_NEXT_ROUND_US must have some real
        // temp_degC for data to be corrected (XScope would say "missing data" if it's too fast)

        #define DO_XSCOPE_INT(name,val,use_at) do {if (use_at == USE_XSCOPE_AT) {xscope_int(name,val);}} while(0)
        //                    ####
        // I have not solved how "unsigned char id" is mapped to "Probe name" in config.xscope
        // but "config.xscope generates macros from probe names" so I guess names are listed in
        // tables as numbers [0..] by some part of the tool

        //      "STARTSTOP" in config.xscope
        #define DO_XSCOPE_START(name,use_at) do {if (use_at == USE_XSCOPE_AT) {xscope_start(name);}} while(0)
        #define DO_XSCOPE_STOP(name,use_at)  do {if (use_at == USE_XSCOPE_AT) {xscope_stop(name);}} while(0)

        // Resolves to two do-while structures since xscope_int also has one

        // In "config.xscope" DISCRETE, which must be handled in offline mode, see
        // "xTIMEcomposer User Guide for tools version 14.0.x" page 133:
        // "Only continous probes can be displayed real-time."
        // If DISCRETE not used here, temps_degC are interpolated (saw tooth curve is not much of temp_degC)
        // Log file is generated as "xscope.xmt" and opens after I stop the RUN

        #define DELAY_FIRST_PRINT_MS 2500 // 2000 seen to fail. 1500 not ok for XSCOPE to work "automagically". Update: 48 kHz problems, 8 kHz ok!
    #else
        #error
    #endif

    // Requires #include <iso646.h> (not tested)
    #define BIT_HIGH(val,iofbit) do { val = val bitor       (1<<iofbit); } while (0)
    #define BIT_LOW(val,iofbit)  do { val = val bitand compl(1<<iofbit); } while (0)
    #define BIT_INV(val,iofbit)  do { val = val xor         (1<<iofbit); } while (0)

    #ifndef DO_DEBUG_1PORTS
        #error DO_DEBUG_1PORTS not defined in Makefile
    #elif (DO_DEBUG_1PORTS == 1)
        #define DEBUG_1PORT_NUM_PULSE_CYCLES 10 // 1 barely visible on my scope
        #define DEBUG_1PORT_HIGH(p)        do { p <: 1; } while (0)
        #define DEBUG_1PORT_LOW(p)         do { p <: 0; } while (0)
        #define DEBUG_1PORT_HIGH_PULSE(p)  do { for (unsigned ix=0; ix<DEBUG_1PORT_NUM_PULSE_CYCLES; ix++) { p <: 1; } p <: 0; } while (0)
        #define DEBUG_1PORT_LOW_PULSE(p)   do { for (unsigned ix=0; ix<DEBUG_1PORT_NUM_PULSE_CYCLES; ix++) { p <: 0; } p <: 1; } while (0)
        #define DEBUG_1PORT_INVERT(p,val)  do { val = val xor 0x01; p <: val; } while (0)
        #define DEBUG_1PORT_VAL(p,val)     do { p <: val; } while (0)
        //
        #define USE_ORANGE_PORT_AT  3 // 0 no place
        #define ORANGE_PORT_AT_1    1 // "standard"
        #define ORANGE_PORT_AT_2    2 // after display
        #define ORANGE_PORT_AT_3    3 // display only
        #define ORANGE_PORT_AT_4    4 // long messages across tiles, needs DEBUG_DO_EXTRA_SYNCH_ACROSS_TILES_FOR_SCOPE==1
        //
        #define DEBUG_1PORT_HIGH_(p,use_at) do {if (use_at == USE_ORANGE_PORT_AT) {p <: 1;}} while (0)
        #define DEBUG_1PORT_LOW_(p,use_at)  do {if (use_at == USE_ORANGE_PORT_AT) {p <: 0;}} while (0)
    #elif (DO_DEBUG_1PORTS == 0)
        #define DEBUG_1PORT_HIGH(p)
        #define DEBUG_1PORT_LOW(p)
        #define DEBUG_1PORT_HIGH_PULSE(p)
        #define DEBUG_1PORT_LOW_PULSE(p)
        #define DEBUG_1PORT_INVERT(p,val)
        #define DEBUG_1PORT_VAL(p,val)
        //
        #define DEBUG_1PORT_HIGH_(p,use_at)
        #define DEBUG_1PORT_LOW_(p,use_at)
    #else
        #error
    #endif

    //      in _Beep_dsp_handlig.xc
    #define SEND_LONG_MESSAGE_ACROSS_TILES_INIT true // true as in product, but large message competes with ch_headset and makes the blipping sound after v0606
                                                     // false removes sending from tile[0] to tile[1] of large messages, removing most of the blipping

    #define DEBUG_DO_EXTRA_SYNCH_ACROSS_TILES_FOR_SCOPE 0 // 0 or 1, which needs ORANGE_PORT_AT_4 to have any meaning

    #define DO_DISPLAY_SCREEN_OFTEN 0 // 0 or 1

    // Compiler crash some times when changing these values, do "Clean project" if so
    #define TEST_HEADSET_BUFFER_TASK 0 // 0 none headset_buffer_task running. v0631 last with more comments:
                                       // 1 combined on tile[0]
                                       // 2 on own logical core on tile[1]
                                       // 3 DEADLOCKS? combined on tile[1]
                                       //   See https://www.xcore.com/viewtopic.php?f=37&t=8244&p=39275#p39275
    //
    #define DEBUG_DSP_ERROR_LOG  0 // 0 or 1
    #define DEBUG_DSP_STATUS_LOG 0 // 0 or 1

    // Cross files. Last tested OK in v0618, all the way including final destination dsp_task_y
    //
    #define DEBUG_MIC_SAMPLES_INCREASING       0 // 0 use mic samples
                                                 // 1 increasing, if so, here's the action:
                                                 //   in verify_mic_samples_delta:
    #define DEBUG_MIC_SAMPLES_INCREASING_CRASH 1 // 0 none of the below (but in-out is on)
                                                 // 1 print
                                                 // 2 xassert
    //      in maths_fix.xc:
    #define DO_Q24_DEBUG_PRINT_IF_MULT_SATURATED 0 // 0 or 1
    #define DO_Q24_TEST_MATH                     0 // 0 or 1
    #define DO_Q24_TEST_CONVERSION               0 // 0 or 1 Last done 24Feb2022

    #define DO_Q24_SET_DEBUG_TAG          0 // 0 or 1
    #define DO_Q24_DEBUG_PRINT_IF_BITPOS 20 // 0 to avoid DO_Q24_DEBUG_PRINT_IF_BITPOS (if 1 above)
    //
    #if (DO_Q24_SET_DEBUG_TAG==1)
        #define SET_DEBUG_TAG(a,b,c,v) set_debug_tag(a,b,c,v)
    #else
        #define SET_DEBUG_TAG(a,b,c,v)
    #endif

    #if ((DO_Q24_DEBUG_PRINT_IF_MULT_SATURATED==1) || (DO_Q24_SET_DEBUG_TAG==1))
        #define SET_DEBUG_TAG_ERROR_PREV const error_info_t error_prev = error
    #else
        #define SET_DEBUG_TAG_ERROR_PREV
    #endif

    #define MEMFILL_TYPE_FOR_LOOP 0
    #define MEMFILL_TYPE_MEMMOVE  1
    #define MEMFILL_TYPE_MEMCPY   2

    #define P_SCOPE_VIOLET_SRC 1 // 0 tile[0]
                                 // 1 tile[1]

    // =====================================================================
    // = PARAMETER INTO LIBRARIES: DO A "Clean Project" after modification =
    // = SPEED_THROUGHPUT_TAG, MAX_WAIT_TIME_TO_SEND_DATA_MS               =
    // =====================================================================
    //
    #define I2C_EXTERNAL_SPEED_K_BITS_PER_SECOND  333 // See box above. 333 is same speed as used in the aquarium in i2c_client_task.xc,
                                                      // i2c_internal_config.clockTicks 300 for older XMOS code struct r_i2c in i2c.h and module_i2c_master
                                                      // 333 writing to the display takes about 33 ms
                                                      // 100 writing to the display takes bout 100 ms
                                                      // BLIP not defendent on this value (tested 100 with v0629)

    #define DEBUG_PRINT_GLOBAL_APP 1 // 0: all printf off
                                     // 1: controlled locally in each xc file. Unit must be connected to xTIMEcomposer. Not for off-line battery usage!
                                     //    DESTROYS ALL TIMING! PRINT FROM ONE CORE DESTROYS FOR ALL CORES!

    #if ((DEBUG_PRINT_GLOBAL_APP == 0) && (USE_XSCOPE == 1))
        #warning XSCOPE probably not working. Set to 1, plus some DO_DEBUG_PRINT must also be set
        // See https://www.teigfam.net/oyvind/home/technology/219-my-beep-brrr-notes/#xscope
    #endif

#else
    #error NESTED .h FILE
#endif /* GLOBALS_H_ */
