/*
 * logger.xc
 *
 *  Created on: 11. feb. 2022
 *      Author: oyvindteig
 */

#define INCLUDES
#ifdef INCLUDES
    #include <xs1.h>
    #include <platform.h> // tile
    #include <timer.h>    // delay_milliseconds(200), XS1_TIMER_HZ etc
    #include <stdint.h>   // uint8_t
    #include <stdio.h>    // printf
    #include <xscope.h>   // xScope logging
    #include <string.h>   // memcpy
    #include <xccompat.h> // REFERENCE_PARAM(my_app_ports_t, my_app_ports) -> my_app_ports_t &my_app_ports
    #include <iso646.h>   // not etc.
    #include <i2c.h>
    #include <i2s.h>
    #include <dsp_qformat.h>
    #include <dsp_fft.h>
    #include <dsp_math.h>
    #include <math.h> // log10 etc
    #include <xassert.h>

    #include "_version.h" // First this..
    #include "_globals.h" // ..then this

    #include "maths.h"
    #include "mic_array.h"
    #include "maths_fix.h" // Needs stdint.h
    #include "mics_in_headset_out.h"
    #include "param.h"
    #include "i2c_client_task.h"
    #include "DAC4_mcp4728.h"
    #include "_Beep_dsp_handling.h"
    #include "port_tile0_task.h"
    #include "logger.h"

    #define DO_DEBUG_PRINT         0 // 0 has consequences if (USE_XSCOPE == 1) and amy timing at all needs to be met
    #define DO_DEBUG_PRINT_SPECIAL 1
    #define DO_WARNINGS            1
    #include "_print_macros_xc.h"
#endif

#if (WARNINGS==1)
    #ifndef USE_XSCOPE
        #error USE_XSCOPE not defined in makefile
    #elif (USE_XSCOPE == 0)
        #warning NO XSCOPE
    #elif ((USE_XSCOPE==1) and (USE_XSCOPE_AT==XSCOPE_AT_2))
        #warning USING XSCOPE AT "STARTSTOP UINT"
    #endif
#endif

// ======================================================================
// ABOUT XSCOPE
// ======================================================================
// xScope timing is for a fill width Window (with 10 parts):
//
// us:  1, 2, 5, 10, 20, 50, 100, 200, 500
// ms:  --"--
// sec: 1, 2, 5, 10, 20, 50
//
// Also observe file "config.xscope" and the xTIMEcomposer User Guide

// xSCOPE beware of OFFSET which for the two first is just a number:
// -----------------------------------------------------------------
// Samples/div |                                    OFFSET
// Always      | Rising   | Falling |               OFFSET
// Auto        | Separate | Persist | Save to png
// Window      | Diff     |                         OFFSET
// -----------------------------------------------------------------
//

// In the xTIME composer User Guide chapter "19xSCOPE performance figures"
// we see that max some 500-900 values per second are picked up by xScope.
// Even if I use XTAG-3 I use an old machine (Mac Mini from 2010)

// So, testing max 500 per second, or 256 inside 500 ms window
// SEEMS LIKE I NEED TO HAVE THE WINDOW ON 1 SEC?


// ======================================================================
// ABOUT I2C
// ======================================================================
// Observe that we have gone to great lengths to store MIC data while the i2C
// is busy to write to the display, see MAX_WAIT_TIME_TO_SEND_DATA_MS. So we do have
// to introduce some load balancing. I'd like the logging_task below to be linear,
// and I'd like it not interfering with mics_in_headset_out_task_a too much by it
// competing with Display_screen

#ifndef I2C_EXTERNAL_SPEED_K_BITS_PER_SECOND
    #error missing
#else
    // DO_DISPLAY_SCREEN_OFTEN 0
    //   per_us  500 tim 25600 late    0  OK 100 ERR 0  ok 25600 err 0 max_us 262
    //   per_us 1000 tim 25344 late    0  OK  99 ERR 0  ok 25344 err 0 max_us 261
    // DO_DISPLAY_SCREEN_OFTEN 1
    //   per_us  300 tim 18936 late 6664  OK 100 ERR 0  ok 25600 err 0 max_us 1253 v0705
    //   per_us  500 tim 18232 late 7368  OK 100 ERR 0  ok 25600 err 0 max_us 1253
    //   per_us 1500 tim 25600 late    0  OK 100 ERR 0  ok 25600 err 0 max_us 1250
    //
    #define PERIODE_TICKS_ ((XS1_TIMER_MHZ * DAC4_REG_FAST_WRITE_COMMAND_NUM_BITS * 1000) / I2C_EXTERNAL_SPEED_K_BITS_PER_SECOND)
            // 100 * 83*1000/333 = 100 * 249 -> 249 us
    #if        (PERIODE_TICKS_ < (XS1_TIMER_MHZ * 500))
        #define PERIODE_TICKS    (XS1_TIMER_MHZ * 273) // 273 * 256 = 69.88 ms scope screen with 5ms/div * 14 = 70 ms
    #else
        #error Find another value
    #endif

    #define XSCOPE_PERIODE_TICKS (XS1_TIMER_MHZ * 1900) // 1.9 ms for NUM_FREQ_POINTS_REAL (256) all in 486 ms
#endif

[[combinable]]
void logger_task (
        chanend                        ch_logger_out_2in, // ready out, curve in
        client i2c_general_commands_if if_i2c_general_commands,
        client port_commands_if        if_port_commands)
{

    timer    tmr;
    time32_t time_ticks;

    int_max_value_t  freq_max_value;
    data_ch_logger_t data_ch_logger;
    bool             ready = true;
    unsigned         debug_curves_cnt = 0;
    unsigned         debug_in_time_cnt = 0;
    unsigned         debug_late_cnt = 0;
    unsigned         debug_OK_cnt = 0;
    unsigned         debug_ERR_cnt = 0;
    unsigned         debug_ok_cnt = 0;
    unsigned         debug_err_cnt = 0;
    unsigned         debug_longest_interval_deltatime_ticks = 0;

    DAC4_context_t DAC4_ctx;
    i2c_uint8_t    i2c_bytes [MUM_DAC4_CHANNELS * 2];
    bool           i2c_error;
    unsigned       iof_curve = 0;
    bool           do_curve = false;

    const unsigned do_scale_ratio_all [MUM_DAC4_CHANNELS] = {true, true, true, true};
    //
    for (unsigned ix=DAC4_CH0_A; ix<MUM_DAC4_CHANNELS; ix++) {
        DAC4_ctx.do_scale_ratio[ix] = do_scale_ratio_all[ix];
    }

    ch_logger_out_2in <: ready;
    debug_print ("logger_task ready%s", "\n");

    if_port_commands.write_port(port_low);

    while (1) {
        select {
            case (do_curve) => tmr when timerafter (time_ticks) :> void : {
                tmr :> time_ticks; // See "timerafter to now and skew" in blog note 165

                DAC4_ctx.curve [DAC4_CH0_A] = data_ch_logger.spectrum[iof_curve];

                xassert (DAC4_ctx.curve [DAC4_CH0_A] <= freq_max_value.val_at_ix);

                DAC4_fill_buffer (DAC4_ctx, i2c_bytes);

                i2c_error = not if_i2c_general_commands.write_reg_ok (
                        I2C_ADDRESS_OF_DAC4,
                        i2c_bytes,
                        sizeof (i2c_bytes));

                if (i2c_error) {
                    debug_err_cnt++;
                    debug_print ("no mcp4728 %x=%x.%x %x.%x %x.%x %x.%x\n",
                            I2C_ADDRESS_OF_DAC4 << I2C_SHIFT_BITS,
                            i2c_bytes[0],i2c_bytes[1],
                            i2c_bytes[2],i2c_bytes[3],
                            i2c_bytes[4],i2c_bytes[5],
                            i2c_bytes[6],i2c_bytes[7]);
                } else {
                    debug_ok_cnt++;
                    debug_print ("ok mcp4728 %x=%x.%x %x.%x %x.%x %x.%x\n",
                            I2C_ADDRESS_OF_DAC4 << I2C_SHIFT_BITS,
                            i2c_bytes[0],i2c_bytes[1],
                            i2c_bytes[2],i2c_bytes[3],
                            i2c_bytes[4],i2c_bytes[5],
                            i2c_bytes[6],i2c_bytes[7]);
                }

                DO_XSCOPE_INT (VALUE, data_ch_logger.spectrum[iof_curve], XSCOPE_AT_2);

                // This may have taken too long. Don't let time_ticks_now suffer:
                //
                time32_t time_ticks_now;
                tmr :> time_ticks_now;

                const unsigned used_deltatime_ticks = (time_ticks_now - time_ticks);
                //
                if (used_deltatime_ticks > debug_longest_interval_deltatime_ticks) {
                    debug_longest_interval_deltatime_ticks = used_deltatime_ticks;
                } else {}

                if (used_deltatime_ticks < PERIODE_TICKS) {
                    time_ticks += PERIODE_TICKS; // Advance linearly
                    debug_in_time_cnt++;
                } else {
                    time_ticks = time_ticks_now; // reset time, causes unavoidable nonlinearity
                    debug_late_cnt++;
                }

                // Below timing cals above since we have a communication here (max was about 128 ms in one case)
                iof_curve++;
                if (iof_curve == NUM_LOGGER_TASK_CURVE_VALUES) { // == 256
                    if_port_commands.write_port(port_low);
                    debug_print ("logger curve done%s", "\n");
                    do_curve = false;
                    DO_XSCOPE_STOP (VALUE, XSCOPE_AT_2);
                    ch_logger_out_2in <: ready;
                }

            } break;

            case ch_logger_out_2in :> data_ch_logger : { // spectrum only has positive values
                ch_logger_out_2in :> freq_max_value;

                debug_curves_cnt++;
                if ((debug_curves_cnt % 100) == 0) {

                    debug_print_special ("per_us %u tim %u late %u  OK %u ERR %u  ok %u err %u max_us %u\n",
                            PERIODE_TICKS/XS1_TIMER_MHZ,
                            debug_in_time_cnt, debug_late_cnt,
                            debug_OK_cnt,      debug_ERR_cnt,
                            debug_ok_cnt,      debug_err_cnt,
                            debug_longest_interval_deltatime_ticks/XS1_TIMER_MHZ);

                    debug_in_time_cnt = 0;
                    debug_late_cnt    = 0;
                    debug_OK_cnt      = 0;
                    debug_ERR_cnt     = 0;
                    debug_ok_cnt      = 0;
                    debug_err_cnt     = 0;
                    //
                    debug_longest_interval_deltatime_ticks = 0;
                } else {
                    debug_print ("logger curve new %u\n", debug_curves_cnt);
                }

                unsigned scale_shift_right_by_bits =  find_downscale_by_bits_param (
                        MAX_Q8_24_IF_POSITIVE_NUM_BITS, // 31
                        DAC4_NUM_BITS,                  // 12
                        freq_max_value.val_at_ix);          // even if signed q8_24 it's never negative

                // This basically finds the scaling factor(s) for the curve (to be used above),
                // plus sets up a synch pulse at DAC4_CH3_D

                const unsigned curve_vals       [MUM_DAC4_CHANNELS] = {0,                         0, 0, SCOPE_TRIGGER_ON_VAL};
                const uint32_t numerator_vals   [MUM_DAC4_CHANNELS] = {1<<DAC4_NUM_BITS,          1, 1, 1};
                const uint32_t denominator_vals [MUM_DAC4_CHANNELS] = {freq_max_value.val_at_ix,      1, 1, 1}; // Arithmetic exception if zero, of course
                const unsigned shift_bits       [MUM_DAC4_CHANNELS] = {scale_shift_right_by_bits, 0, 0, 0};
                //
                for (unsigned ix=DAC4_CH0_A; ix<MUM_DAC4_CHANNELS; ix++) {
                    DAC4_ctx.curve                     [ix]     = curve_vals         [ix];
                    // DAC4_ctx.do_scale_ratio         [ix]     = do_scale_ratio_all [ix]; done
                    DAC4_ctx.scale_ratio               [ix].num = numerator_vals     [ix];
                    DAC4_ctx.scale_ratio               [ix].den = denominator_vals   [ix];
                    DAC4_ctx.scale_shift_right_by_bits [ix]     = shift_bits         [ix];
                }

                xassert (freq_max_value.val_at_ix > 0); // To avoid crash quite soon

                DAC4_fill_buffer (DAC4_ctx, i2c_bytes);

                if_port_commands.write_port(port_high);

                i2c_error = not if_i2c_general_commands.write_reg_ok (
                        I2C_ADDRESS_OF_DAC4,
                        i2c_bytes,
                        sizeof (i2c_bytes));

                if (i2c_error) {
                    debug_ERR_cnt++;
                    debug_print ("NO mcp4728 %x=%x.%x %x.%x %x.%x %x.%x\n",
                            I2C_ADDRESS_OF_DAC4 << I2C_SHIFT_BITS,
                            i2c_bytes[0],i2c_bytes[1],
                            i2c_bytes[2],i2c_bytes[3],
                            i2c_bytes[4],i2c_bytes[5],
                            i2c_bytes[6],i2c_bytes[7]);
                } else {
                    debug_OK_cnt++;
                    debug_print ("OK mcp4728 %x=%x.%x %x.%x %x.%x %x.%x\n",
                            I2C_ADDRESS_OF_DAC4 << I2C_SHIFT_BITS,
                            i2c_bytes[0],i2c_bytes[1],
                            i2c_bytes[2],i2c_bytes[3],
                            i2c_bytes[4],i2c_bytes[5],
                            i2c_bytes[6],i2c_bytes[7]);
                }

                DAC4_ctx.curve [DAC4_CH3_D] = SCOPE_TRIGGER_OFF_VAL;

                DO_XSCOPE_START (VALUE, XSCOPE_AT_2);
                do_curve = true;
                iof_curve = 0;
                tmr :> time_ticks; // reset time
                time_ticks += PERIODE_TICKS; // since we have done a if_port_commands.write_port here
            } break;
        }
    }
}

