/*
 * maths_fix.xc
 *
 *  Created on: 5. nov. 2021
 *      Author: oyvindteig
 */

// See more discussiion in the header file

// OBSERVE that lib_dsp dsp_math.h + dsp_math.xc have all that's needed here I assume

#define INCLUDES
#ifdef INCLUDES
    #include <xs1.h>
    #include <platform.h> // tile
    #include <timer.h>    // delay_milliseconds(200), XS1_TIMER_HZ etc
    #include <stdint.h>   // uint8_t
    #include <stdio.h>    // printf
    #include <xccompat.h> // REFERENCE_PARAM(my_app_ports_t, my_app_ports) -> my_app_ports_t &my_app_ports
    #include <iso646.h>   // not etc.
    #include <math.h>     // HUGE_VAL not resolved
    #include <dsp_math.h>
    #include <dsp_qformat.h>
    #include <xassert.h>

    #include "_version.h" // First this..
    #include "_globals.h" // ..then this
    #include "maths_fix.h"

    #define DO_DEBUG_PRINT         1 // 0 has consequences if (USE_XSCOPE == 1) and amy timing at all needs to be met
    #define DO_DEBUG_PRINT_SPECIAL 0
    #define DO_WARNINGS            1
    #include "_print_macros_xc.h"
#endif


// Wikipedia calls the parts "integer" and "fraction".
// The _INT name here would cover both signed and unsigned (as said above)
//
#define Q_FORMAT_MULTIPLIER_FRACS_6 1000000L // For q8_24 and uq8_24 this give 6 figures after the decimal point

// Used to develop Q_FORMAT_FRAC:
// signed int         val_int = value / (1<<24);
// unsigned long long ll_frac = (value bitand ((1<<24)-1));
//                    ll_frac = ll_frac * 1000000L;
// unsigned int       frac    = (unsigned int) (ll_frac / (1<<24));
//
// REQUIRES USE OF sprintf_fraction_24bit (or using sprintf and %06u with leading zeroes) or
// if not lib_logging (none set to 1: DEBUG_PRINT_ENABLE_ALL, DEBUG_PRINT_ENABLE) then %6u directly in debug_print
//
#define Q_FORMAT_FRAC_(val,num_frac_bits,mul)       \
            ((unsigned int)                         \
             (                                      \
              (                                     \
               (                                    \
                (unsigned long long)                \
                (val bitand ((1<<num_frac_bits)-1)) \
               ) * mul                              \
              ) / (1<<num_frac_bits)                \
             )                                      \
            )

// A negative 2's complement number does not have the same view of each bit as a positive number
// -1 is only FF while +1 is only one 1! Therefore, imvert them! This took me days..
//
#define Q_FORMAT_FRAC(val,num_frac_bits,mul) \
            (val>=0) \
                    ? Q_FORMAT_FRAC_(val,num_frac_bits,mul) \
                    : Q_FORMAT_FRAC_(-val,num_frac_bits,mul)

#define Q_FORMAT_FLOAT(val,num_frac_bits) ((float)val)/(1<<num_frac_bits)


// ABOUT %06u AND USE OF sprintf in uq24_unsigned_to_str and q24_signed_to_str
//
// With one of these set to 1: DEBUG_PRINT_ENABLE_ALL or DEBUG_PRINT_ENABLE (signifying lib_logging) then
// "%06u" is not handled in debug_print, but sprintf expands "%06u" ok.
// In other words, if lib_logging is NOT used, then sprintf_fraction_24bit for %06u is not needed
//

signed uq24_unsigned_to_str (          // returns num_chars
        char           chars [],       // Includes \0 character
        const uint32_t value_unsigned) // always unsigned
{
    int num_chars; // I guess, negative if some conversion error
    uint32_t fraction_unsigned = Q_FORMAT_FRAC (value_unsigned, 24, Q_FORMAT_MULTIPLIER_FRACS_6);

    chars[0] = 0;

    num_chars = sprintf (chars, "%u.%06u", Q_FORMAT_INT(value_unsigned,24), fraction_unsigned);

    return num_chars;
}

signed q24_signed_to_str (          // returns num_chars
        char          chars [],     // Includes \0 character
        const int32_t value_signed) // always signed
{
    int num_chars; // I guess, negative if some conversion error

    chars[0] = 0;

    // if (value == MIN_Q8_24) .. instead, going via value_signed_64 below
    if (value_signed < 0) {

        uint32_t fraction_unsigned = Q_FORMAT_FRAC (value_signed, 24, Q_FORMAT_MULTIPLIER_FRACS_6);

        // Set "-" by hand (and avoid it being set automatically by inverting the value for Q_FORMAT_INT),
        // since in the case for negative value with integer part zero, like -0.25 then that
        // minus is not shown with standard handling:
        
        // Via 64 bits since negating MIN_Q8_24 (0x80000000)(dec.-2147483648) is meaningless
        // since MAX_Q8_24 is not the negated of MIN_Q8_24 ("it's one more" = out of range)
        // i.e. it becomes -4294967168
        
        uint64_t value_signed_64 = (uint64_t) -value_signed;
        uint32_t value_signed_32 = (uint32_t)  value_signed_64;
        //
        num_chars = sprintf (chars, "-%u.%06u", Q_FORMAT_INT(value_signed_32,24), fraction_unsigned);
    } else {
        num_chars = uq24_unsigned_to_str (chars, (unsigned) value_signed);
    }

    return num_chars;
}

// Called if DO_Q24_SET_DEBUG_TAG==1
// In that case used as SET_DEBUG_TAG
//
bool set_debug_tag ( 
          error_info_t       &error, // io
          const error_info_t error_prev,
          const unsigned     tag_bitpos, // 0 to avoid DO_Q24_DEBUG_PRINT_IF_BITPOS
          const signed       value) 
{
    bool triggered = false;
    
    if (error_prev.err_cnt != error.err_cnt) {
        
        error.debug_tag or_eq (1 << tag_bitpos);
        
        #if (DO_Q24_DEBUG_PRINT_IF_BITPOS != 0)
            if (DO_Q24_DEBUG_PRINT_IF_BITPOS == tag_bitpos) {
                debug_print ("Tag[pos:%u val: 0x%X] Val: %d 0x%X\n", tag_bitpos, error.debug_tag, value, value);
                triggered = true;
            } else {}
        #endif

    } else {}
    
    return triggered;
}


int32_t my_dsp_math_multiply_sat ( // returns product_32_sat
        int32_t        input1_value,
        int32_t        input2_value,
        const int32_t  q_format,
        error_info_t   &error) {

    int32_t product_32_sat;
    int32_t product_32;

    // lsats and lextract also exist as macros (se xTIMEcomposer User Guide)

    // Assembler asm:
    // maccs    Mulitply and accumulate signed
    // lsats    Saturate signed (xCORE-200 only)
    // lextract Bitfield extraction from register pair (xCORE-200 only)
    //
    product_32_sat = dsp_math_multiply_sat (input1_value, input2_value, q_format);

    // Assembler asm:
    // maccs    Mulitply and accumulate signed
    // lextract Bitfield extraction from register pair (xCORE-200 only)
    //
    product_32 = dsp_math_multiply (input1_value, input2_value, q_format);

    // In an earlier version of my_dsp_math_multiply_sat (last in v0510)
    // I did a local 64-bit = 32-bit * 32-bit then shift >> 24. In that case
    // the rounding to 32 bits was in some cases different by one bit value
    // like FF65DF66 vs FF65DF65. I then decided to just compare the sat and
    // non-sat versions from the library, where the sat only has in intermediate
    // lsats instruction. This also took some time to see!

    if (product_32_sat != product_32) {
        SET_DEBUG_TAG_ERROR_PREV;
        
        error.err_cnt++;
        
        SET_DEBUG_TAG (error, error_prev, 20, product_32);
        
        // product_32_sat is different meaning it has been saturated
        #if (DO_Q24_DEBUG_PRINT_IF_MULT_SATURATED==1)
            char q24_1_str [Q_FORMAT_STR_LEN];
            char q24_2_str [Q_FORMAT_STR_LEN];
            char q24_3_str [Q_FORMAT_STR_LEN];
            char q24_4_str [Q_FORMAT_STR_LEN];

            q24_signed_to_str (q24_1_str,input1_value);
            q24_signed_to_str (q24_2_str,input2_value);
            q24_signed_to_str (q24_3_str,product_32_sat);
            q24_signed_to_str (q24_4_str,product_32);

            debug_print ("tag2[%X]: %d * %d = %d (32:%#x) != %d (32s:%#x)\nSAT: %s %s %s %s\n",
                    error.debug_tag,
                    input1_value,    
                    input2_value,   
                    product_32_sat,
                    product_32_sat,
                    product_32,    
                    product_32,   
                    q24_1_str, // input1_value 
                    q24_2_str, // input2_value
                    q24_3_str, // product_32_sat
                    q24_4_str);// product_32
        #endif

        // Not setting product_32_sat to Q24_NaN, because we have incremented err_cnt,
        // and we have a saturated value to try to work with

    } else {}

    return product_32_sat;
}

// Multiply not allowed to overflow/underflow into highest 32 bits word
//
int32_t my_math_multiply_32bit_sat ( // returns product_32_sat
        const int32_t value,
        const int32_t mul_N,
        error_info_t  &error) 
{
    int64_t product_64; 
    int32_t product_32;
    
    SET_DEBUG_TAG_ERROR_PREV;
    
    product_64 = (int64_t) value * (int64_t) mul_N;
 
    if (product_64 >= 0L) {
        if (product_64 > INT32_MAX) { 
            // something in higher word
            product_32 = INT32_MAX;
            error.err_cnt++;
        } else {
            product_32 = (int32_t) product_64;
        } 
    } else {
        if (product_64 < INT32_MIN) {
            // something in higher word
            product_32 = INT32_MIN;
            error.err_cnt++;
        } else {
            product_32 = (int32_t) product_64;
        }
    }
    
    SET_DEBUG_TAG (error, error_prev, 21, value);
    
    #if (DO_Q24_DEBUG_PRINT_IF_MULT_SATURATED==1)  
        if (error_prev.err_cnt != error.err_cnt) {
            debug_print ("tag3[%X] value:%d mul:%d sat:%d",
                    error.debug_tag,
                    value,
                    mul_N,
                    product_32);
        } else {}
    
    #endif
    
    return (int32_t) product_64;               
}


void test_ln_of_1 () {

    // Setting value to 1 in uq8_24 is 1/(2exp24) = 1/16777216
    // from macOS Calculator:
    //       = 0,000000059604645
    //     ln (0,000000059604645) = −16,635532333438687
    
    // Printed out here:
    // ln(1):-279097920 -16.635532 vs:-279097914
    //        Q24_LN_MIN              Q24_LN_MIN_

    q8_24 ln_signed = dsp_math_log (1); // ln

    char ln_str [Q_FORMAT_STR_LEN];

    q24_signed_to_str (ln_str, ln_signed);
    debug_print ("ln(1):%d %s vs:%d\n", ln_signed, ln_str, Q24_LN_MIN_);
}


// Calculate dB (decibel) of unsigned uq8_24
//
q8_24 my_dB (
        const uq8_24 unsigned_operand,
        error_info_t &error)
{
    q8_24 dB_signed; // signed/unsigned rather important here, using such names to get it right
    
    error.debug_tag bitand compl (1<<20); // qwe clear bit

    const error_info_t error_prev = error;

    if (unsigned_operand > MAX_Q8_24) { // > 0x7FFFFFFF, ie from UQ24_128 0x80000000

        // To avoid  dsp_math_log being called with anything else than positive range of the signed [0..127.999],
        // since it directly delivers an unsigned into a signed!
        //
        // Letting it pass without this test might cause crash on arithmetic overflow with division by zero in
        // log2_with_remainder at z = dsp_math_divide(y, zden, 24); (when UQ24_128 0x80000000 is passed to it)

        // q8_24 dsp_math_log     (uq8_24 x) {
        //                         !!!!!!
        //                         !!!!!! This must be a coding error in lib_dsp
        //                         !!!!!!
        //    log2_with_remainder (q8_24 x,..)
        //
        // In this case this also covers log(negative number) = NaN for any log base
        error.err_cnt++;
        SET_DEBUG_TAG (error, error_prev, 11, unsigned_operand);
    } else {
        q8_24 log_e_signed;
        
        if (unsigned_operand == 0) { 
            // =< meaningless since unsigned_operand is unsigned
            // unsigned_operand is the magnitude or sqrt: zero is indeed possible
            // No need to calculate this when I know the answer from test_ln_of_1 or Q24 macro
            //        
            log_e_signed = Q24_LN_MIN; // -16.635532 = -279097920
        } else {
            log_e_signed = dsp_math_log (unsigned_operand); // ln(unsigned_operand). This calls
                                                            // log2_with_remainder(unsigned q8_24)
                                                            // so unsigned_operand is max just below 128
        }

        // Since dsp_math_multiply is signed, we need to map to q8_24, therefore according to my calculations then
        // 0x0B504F33 =  11,313708484172823 is max since it squared and >> 24 is
        // 0x7FFFFFFA = 127,999999940395355
        //
        q8_24 log_10_signed;

        // log_e_signed −2,657825588340702
        log_10_signed = my_dsp_math_multiply_sat (log_e_signed,  Q24_LOG10_OF_E_NUM, 24, error); // log10(operand) = ln(operand) * log10(e)       
        SET_DEBUG_TAG (error, error_prev, 12, log_10_signed);
        
        // log_10_signed = -7,224719 so many times and always here, causes underflow v0615
        // -7,224719 -121210681 (0xF8C678C7)
        // The first overflow is in the below my_dsp_math_multiply_sat
        // SET_DEBUG_TAG (error, error_prev, 20, product_32);
        dB_signed     = my_dsp_math_multiply_sat (log_10_signed, Q24(20),   24, error); // dB_signed(operand)    = 20 * log10(operand)
        
        SET_DEBUG_TAG (error, error_prev, 13, dB_signed);

        if ((error.debug_tag bitand (1<<20)) != 0) {
            char q24_1_str [Q_FORMAT_STR_LEN];
            char q24_2_str [Q_FORMAT_STR_LEN];
            char q24_3_str [Q_FORMAT_STR_LEN];
            char q24_4_str [Q_FORMAT_STR_LEN];

            uq24_unsigned_to_str (q24_1_str,unsigned_operand);
            q24_signed_to_str    (q24_2_str,log_e_signed);
            q24_signed_to_str    (q24_3_str,log_10_signed);
            q24_signed_to_str    (q24_4_str,dB_signed);
            debug_print ("tag4[%X] uo:%u \"%s\"  les:%d \"%s\"  l10s:%d \"%s\"  dBs:%d \"%s\"\n",
                    error.debug_tag,
                    unsigned_operand, q24_1_str,
                    log_e_signed,     q24_2_str,
                    log_10_signed,    q24_3_str,
                    dB_signed,        q24_4_str);
            /*debug_print ("tag4[%X] uo:%u\tles:%d\tl10s:%d\tdBs:%d\n",
                       error.debug_tag,
                       unsigned_operand,
                       log_e_signed,     
                       log_10_signed, 
                       dB_signed);
            debug_print ("tag4[%X] uo:%s\tles:%s\tl10s:%s\tdBs:%s\n",
                    error.debug_tag,
                    q24_1_str,
                    q24_2_str,
                    q24_3_str,
                    q24_4_str); */
        } else {}
    }

    if (error.err_cnt != error_prev.err_cnt) {
        // log(0) = NaN for any log base
        // log(negative number) = NaN for any log base
        // any two my_dsp_math_multiply_sat calls with overflow
        //
        dB_signed = Q24_NaN;
    } else {}

    return dB_signed;
}  // my_dB

// Use like this:
/*
scale_shift_right_by_bits = find_downscale_by_bits_param (
        MAX_Q8_24_IF_POSITIVE_NUM_BITS, // 31
        DAC4_NUM_BITS,                  // 12
        freq_max_value.val_at_ix);        // even if signed q8_24 it's never negative
*/
//
// PRO: avoids using int32_ratio_t or uint32_ratio_t and involving 64 bits values,
// CON: does not autoscale to full range
//
// OBSERVE: NOT VERIFIED!
//
unsigned find_downscale_by_bits_param (
    const unsigned max_range_num_bits, // 31
    const unsigned scaled_to_num_bits, // 12
    unsigned       val_positive) // positive but packed in unsigned (for "math_dsp reasons")
{
     unsigned scale_shift_right_by_bits;
     bool mask_not_set = true;
     unsigned bit_no = max_range_num_bits-1; // 30 is first = 0x40000000

     if (val_positive == 0) {
         bit_no = 0;
     } else {
         while (mask_not_set) {
             if ((val_positive bitand (1<<bit_no)) != 0) {
                 mask_not_set = false; // finished
             } else {
                 if (bit_no == 0) {
                     mask_not_set = false; // zero already taken care of, but anyhow
                 } else {
                     bit_no--;
                 }
             }
         }
     }

     if (bit_no >= scaled_to_num_bits) { // If bit12 or [12] is set, it's one too high ([0..11] is range)
         scale_shift_right_by_bits = (bit_no - (scaled_to_num_bits-1)); // 12 - (12-1) = 1
     } else {
         scale_shift_right_by_bits = 0; // value is within range, nothing to do
     }

     return scale_shift_right_by_bits;
}

uint32_t downscale_by_uint32_ratio (
        const uint32_t       value,
        const uint32_ratio_t ratio) {

    uint32_t product_32;

    if ((ratio.num == 1) and (ratio.den == 1)) {
        product_32 = value;
    } else {
        uint64_t product_64; // Overflow/underflow handled later with xassert on total result

        product_64 = ( ((uint64_t) value) * ((uint64_t) ratio.num) );
        product_64 /=                        (uint64_t) ratio.den;

        product_32 = (uint32_t) product_64;
    }

    return product_32;
}

 #if (DO_Q24_TEST_MATH==1)
    void q24_test_math_iff (void)
    {
        #define NUM_VALS 13

        const uq8_24 uq8_24_nums_unsigned [NUM_VALS] = {
                Q24(0.25), Q24(0.5), Q24(1), Q24(2), Q24(4), Q24(8), Q24(10), Q24(16), Q24(32), Q24(64), Q24(100),
                UQ24_128,  // 0x80000000
                UQ24_MAX}; // 0xFFFFFFFF

        uq8_24 uq8_24_sqrt_unsigned [NUM_VALS];
        q8_24  q8_24_dB_signed      [NUM_VALS];

        error_info_t error;

        char  nums_str [Q_FORMAT_STR_LEN];
        char  sqrt_str [Q_FORMAT_STR_LEN];
        char  dB_str   [Q_FORMAT_STR_LEN];

        for (unsigned ix=0; ix<NUM_VALS; ix++) {
            {
                uq24_unsigned_to_str (nums_str, uq8_24_nums_unsigned[ix]);
                debug_print ("[%u]\tnums:%s", ix, nums_str);
            }

            uq8_24_sqrt_unsigned[ix] = dsp_math_sqrt (uq8_24_nums_unsigned[ix]);
            {
                uq24_unsigned_to_str (sqrt_str, uq8_24_sqrt_unsigned[ix]);
                debug_print ("\tsqrt:%s", sqrt_str);
            }

            error.err_cnt = 0;
            error.val_aux_cnt = 0;

            q8_24_dB_signed[ix] = my_dB (uq8_24_nums_unsigned[ix], error);
            // log2_with_remainder (
            {
                q24_signed_to_str (dB_str, q8_24_dB_signed[ix]);
                debug_print ("\tdB:%s ok:%u err:%u\n", dB_str, error.val_aux_cnt, error.err_cnt);
            }
        }
        /* v0514
        [0] nums:0.250000   sqrt:0.500000   dB:-12.041199 ok:2 err:0
        [1] nums:0.500000   sqrt:0.707106   dB:-6.020599 ok:2 err:0
        [2] nums:1.000000   sqrt:1.000000   dB:0.000000 ok:2 err:0
        [3] nums:2.000000   sqrt:1.414213   dB:6.020599 ok:2 err:0
        [4] nums:4.000000   sqrt:2.000000   dB:12.041199 ok:2 err:0
        [5] nums:8.000000   sqrt:2.828427   dB:18.061798 ok:2 err:0
        [6] nums:10.000000  sqrt:3.162277   dB:19.999998 ok:2 err:0
        [7] nums:16.000000  sqrt:4.000000   dB:24.082398 ok:2 err:0
        [8] nums:32.000000  sqrt:5.656854   dB:30.102998 ok:2 err:0
        [9] nums:64.000000  sqrt:8.000000   dB:36.123597 ok:2 err:0
        [10]    nums:100.000000 sqrt:10.000000  dB:39.999998 ok:2 err:0
        [11]    nums:128.000000 sqrt:11.313708  dB:0.000000 ok:0 err:1
        [12]    nums:255.999999 sqrt:15.999999  dB:0.000000 ok:0 err:1
        */
    }
#endif

#if (DO_Q24_TEST_CONVERSION==1)

    // With this scheme, any -0.nn value will ned have the negative sign if not this is used (-1 is 0xffffffff):
    // Obsoleted, but used in q24_test_conversion_iff
    //
    #define Q_FORMAT_SIGN_FOR_MINUSZEROPOINT_STR(val_signed,num_frac_bits) ((val_signed>>num_frac_bits)==(-1)) ? "-" : ""

    // Obsoleted, but used in q24_test_conversion_iff
    //
    #define Q_FORMAT_CHAR_LEN_FRACS_6 7  // Space for \0 needed

    signed sprintf_fraction_24bit (  // returns num_chars
            char           chars [], // Includes \0 character
            const uint32_t fraction) // always unsigned
    {
        int num_chars; // I guess, negative if some conversion error
        //                           %06u" see this described elsewhere
        num_chars = sprintf (chars, "%06u", fraction); // Observe Q_FORMAT_MULTIPLIER_FRACS_6

        return num_chars;
    }

    void q24_test_conversion_iff (void)
    {
        timer    tmr;
        time32_t time_start_dsp_ticks;
        time32_t time_end_dsp_ticks;

        uq8_24 test_sqrt = MAX_Q8_24; // 0x7FFFFFFF = 127.999999940395355224609375 or 128.0
        uq8_24 sqrt;                  // 11.313708496350582 or 11.313708
        char q24_fraction_str [Q_FORMAT_CHAR_LEN_FRACS_6]; // Observe use %s since string

        tmr :> time_start_dsp_ticks;
        sqrt = dsp_math_sqrt (test_sqrt);
        tmr :> time_end_dsp_ticks;

        debug_print_full ("F1 %f root = %f (t:%u)\n", Q_FORMAT_FLOAT(test_sqrt,24),
                Q_FORMAT_FLOAT(sqrt,24),
                time_end_dsp_ticks - time_start_dsp_ticks);

        q8_24 value_signed = 0x7FFFFFFF; // MAX_Q8_24
        debug_print_full ("F2 q8_24 of %#x is %f\n", value_signed, Q_FORMAT_FLOAT(value_signed,24));
        //                 q8_24 of 0x7FFFFFFF is       128.000000 (1 part below 128 would have been correct)
        //             if uq8_24 then it correctly does 127.999999

        value_signed = 0x80000000; // MIN_Q8_24
        debug_print_full ("F3 q8_24 of %#x is %f\n", value_signed, Q_FORMAT_FLOAT(value_signed,24));
        //                 q8_24 of 0x80000000 is -128.000000

        value_signed = 105414357>>1; // 3243F6A PI_Q8_24 (PI2_Q8_24>>1)
        debug_print_full ("F4 q8_24 of %#x is %f\n", value_signed, Q_FORMAT_FLOAT(value_signed,24));
        //                 q8_24 of 3243F6A is 3.141593
        //                          3 243F6A  hex
        //                          3 2375530 dec and pow(2,24) = 16777216 dec
        //                            2375530 / 16777216 = 0,141592621803284 dec

        sprintf_fraction_24bit (q24_fraction_str, Q_FORMAT_FRAC (value_signed, 24, Q_FORMAT_MULTIPLIER_FRACS_6));
        debug_print ("I1 q8_24 of %#x is %d.%s\n", value_signed, Q_FORMAT_INT(value_signed,24), q24_fraction_str);

        value_signed = 0x80000000; // MIN_Q8_24 -128.0
        sprintf_fraction_24bit (q24_fraction_str, Q_FORMAT_FRAC (value_signed, 24, Q_FORMAT_MULTIPLIER_FRACS_6));
        debug_print ("I2 q8_24 of %#x is %d.%s\n", value_signed, Q_FORMAT_INT(value_signed,24), q24_fraction_str);

        uq8_24 value_unsigned = 0xFFFFFFFF; // UQ24_MAX 255.999999 or 256
        sprintf_fraction_24bit (q24_fraction_str, Q_FORMAT_FRAC (value_unsigned, 24, Q_FORMAT_MULTIPLIER_FRACS_6));
        debug_print ("I3 q8_24 of %#x is %d.%s\n", value_unsigned, Q_FORMAT_INT(value_unsigned,24), q24_fraction_str);

        debug_print_full ("F5 q8_24 of %#x is %f\n", value_unsigned, Q_FORMAT_FLOAT(value_unsigned,24));

        value_unsigned = 0; // UQ24_MIN
        sprintf_fraction_24bit (q24_fraction_str, Q_FORMAT_FRAC (value_unsigned, 24, Q_FORMAT_MULTIPLIER_FRACS_6));
        debug_print ("I4 q8_24 of %#x is %d.%s\n", value_unsigned, Q_FORMAT_INT(value_unsigned,24), q24_fraction_str);

        //               ### observe this rather nice macro (needs "dsp_qformat.h")
        value_unsigned = Q24(1.001122); // prints 1.001121 (1 as last)
        sprintf_fraction_24bit (q24_fraction_str, Q_FORMAT_FRAC (value_unsigned, 24, Q_FORMAT_MULTIPLIER_FRACS_6));
        debug_print ("I5 q8_24 of %#x is %d.%s\n", value_unsigned, Q_FORMAT_INT(value_unsigned,24), q24_fraction_str);

        // prints 1.001122
        debug_print_full ("F6 q8_24 of %#x is %f\n", value_unsigned, Q_FORMAT_FLOAT(value_unsigned,24));

        value_unsigned = Q24(0.25);
        sprintf_fraction_24bit (q24_fraction_str, Q_FORMAT_FRAC (value_unsigned, 24, Q_FORMAT_MULTIPLIER_FRACS_6));
        debug_print ("I6 q8_24 of %#x is %d.%s\n", value_unsigned, Q_FORMAT_INT(value_unsigned,24), q24_fraction_str);

        value_signed = Q24(-0.25);
        sprintf_fraction_24bit (q24_fraction_str, Q_FORMAT_FRAC (value_signed, 24, Q_FORMAT_MULTIPLIER_FRACS_6));
        debug_print ("I7 q8_24 of %#x is %s%d.%s\n",
                value_signed,
                Q_FORMAT_SIGN_FOR_MINUSZEROPOINT_STR(value_signed,24),
                Q_FORMAT_INT(value_signed,24),
                q24_fraction_str);

        value_signed = Q24(-1.25);
        sprintf_fraction_24bit (q24_fraction_str, Q_FORMAT_FRAC (value_signed, 24, Q_FORMAT_MULTIPLIER_FRACS_6));
        debug_print ("I8 q8_24 of %#x is %s%d.%s\n",
                value_signed,
                Q_FORMAT_SIGN_FOR_MINUSZEROPOINT_STR(value_signed,24),
                Q_FORMAT_INT(value_signed,24),
                q24_fraction_str);

        // The above shows the process of how I reached the below:

        char q24_str [Q_FORMAT_STR_LEN]; // Observe use %s since string

        value_signed = Q24(-0.25);
        q24_signed_to_str (q24_str,value_signed);
        debug_print ("I9 q8_24 of %#x is %s\n", value_signed, q24_str);

        value_signed = Q24(-1.25);
        q24_signed_to_str (q24_str,value_signed);
        debug_print ("I10 q8_24 of %#x is %s\n", value_signed, q24_str);

        value_signed = Q24(2.25);
        q24_signed_to_str (q24_str,value_signed);
        debug_print ("I11 q8_24 of %#x is %s\n", value_signed, q24_str);

        value_unsigned = Q24(2.25);
        uq24_unsigned_to_str (q24_str,value_unsigned);
        debug_print ("I12 uq8_24 of %#x is %s\n", value_unsigned, q24_str);
        
        value_unsigned = Q24(0);
        uq24_unsigned_to_str (q24_str,value_unsigned);
        debug_print ("I13 uq8_24 of %#x is %s\n", value_unsigned, q24_str);
        
        value_signed = Q24(0);
        q24_signed_to_str (q24_str,value_signed);
        debug_print ("I14 q8_24 of %#x is %s\n", value_signed, q24_str);
                
        value_signed = -121210681; // F8C678C7, since that number was problematic in my_dB string conversion ("000000")
        q24_signed_to_str (q24_str,value_signed);
        debug_print ("I15 q8_24 of %#x is %s\n", value_signed, q24_str);
                
        value_signed = 0x80000000; // MIN_Q8_24 (0x80000000) -2147483648;
        q24_signed_to_str (q24_str,value_signed);
        debug_print ("I16 q8_24 of %#x is %s\n", value_signed, q24_str);

        /* If none set to 1: DEBUG_PRINT_ENABLE_ALL, DEBUG_PRINT_ENABLE (in Makefile) (standard printf)
         * Only the intermingled float prints shwon here
         *
        F1 128.000000 root = 11.313708 (t:58)
        F2 q8_24 of 0x7fffffff is 128.000000
        F3 q8_24 of 0x80000000 is -128.000000
        F4 q8_24 of 0x3243f6a is 3.141593
        F5 q8_24 of 0xffffffff is 256.000000
        F6 q8_24 of 0x1004988 is 1.001122
        */

        /* If either set to 1: DEBUG_PRINT_ENABLE_ALL, DEBUG_PRINT_ENABLE using lib_logging (in Makefile) (floats gone)
         *
        I1 q8_24 of 3243F6A is 3.141592
        I2 q8_24 of 80000000 is -128.000000
        I3 q8_24 of FFFFFFFF is 255.999999
        I4 q8_24 of 0 is 0.000000
        I5 q8_24 of 1004988 is 1.001121
        I6 q8_24 of 400000 is 0.250000
        I7 q8_24 of FFC00000 is -0.250000
        I8 q8_24 of FEC00000 is -1.250000
        I9 q8_24 of FFC00000 is -0.250000
        I10 q8_24 of FEC00000 is -1.250000
        I11 q8_24 of 2400000 is 2.250000
        I12 uq8_24 of 2400000 is 2.250000
        I13 uq8_24 of 0 is 0.000000
        I14 q8_24 of 0 is 0.000000
        I15 q8_24 of F8C678C7 is -7.224719
        I16 q8_24 of 80000000 is -128.000000
        */
    }
#endif
